@@ -12,16 +12,25 @@
 # See the License for the specific language governing permissions and
 # limitations under the License.
 
+from __future__ import annotations
+
 import argparse
 import functools
+import itertools
 import os
 import sys
-from typing import Any, List, Optional, Sequence, Union
+from collections.abc import Iterable
+from typing import Any, Callable, Literal, Sequence
+
+import argcomplete
 
 from nox import _option_set
 from nox.tasks import discover_manifest, filter_manifest, load_nox_module
+from nox.virtualenv import ALL_VENVS
 
-"""All of nox's configuration options."""
+ReuseVenvType = Literal["no", "yes", "never", "always"]
+
+"""All of Nox's configuration options."""
 
 options = _option_set.OptionSet(
     description="Nox is a Python automation toolkit.", add_help=False
@@ -46,7 +55,8 @@ options.add_groups(
     _option_set.OptionGroup(
         "environment",
         "Environment options",
-        "These arguments are used to control Nox's creation and usage of virtual environments.",
+        "These arguments are used to control Nox's creation and usage of virtual"
+        " environments.",
     ),
     _option_set.OptionGroup(
         "execution",
@@ -61,11 +71,11 @@ options.add_groups(
 )
 
 
-def _sessions_and_keywords_merge_func(
+def _sessions_merge_func(
     key: str, command_args: argparse.Namespace, noxfile_args: argparse.Namespace
-) -> List[str]:
-    """Only return the Noxfile value for sessions/keywords if neither sessions
-    or keywords are specified on the command-line.
+) -> list[str]:
+    """Only return the Noxfile value for sessions/keywords if neither sessions,
+    keywords or tags are specified on the command-line.
 
     Args:
         key (str): This function is used for both the "sessions" and "keywords"
@@ -75,15 +85,19 @@ def _sessions_and_keywords_merge_func(
             command-line.
         noxfile_Args (_option_set.Namespace): The options specified in the
             Noxfile."""
-    if not command_args.sessions and not command_args.keywords:
-        return getattr(noxfile_args, key)
-    return getattr(command_args, key)
+    if (
+        not command_args.sessions
+        and not command_args.keywords
+        and not command_args.tags
+    ):
+        return getattr(noxfile_args, key)  # type: ignore[no-any-return]
+    return getattr(command_args, key)  # type: ignore[no-any-return]
 
 
 def _default_venv_backend_merge_func(
     command_args: argparse.Namespace, noxfile_args: argparse.Namespace
 ) -> str:
-    """Merge default_venv_backend from command args and nox file. Default is "virtualenv".
+    """Merge default_venv_backend from command args and Noxfile. Default is "virtualenv".
 
     Args:
         command_args (_option_set.Namespace): The options specified on the
@@ -101,7 +115,7 @@ def _default_venv_backend_merge_func(
 def _force_venv_backend_merge_func(
     command_args: argparse.Namespace, noxfile_args: argparse.Namespace
 ) -> str:
-    """Merge force_venv_backend from command args and nox file. Default is None.
+    """Merge force_venv_backend from command args and Noxfile. Default is None.
 
     Args:
         command_args (_option_set.Namespace): The options specified on the
@@ -117,10 +131,8 @@ def _force_venv_backend_merge_func(
             raise ValueError(
                 "You can not use `--no-venv` with a non-none `--force-venv-backend`"
             )
-        else:
-            return "none"
-    else:
-        return command_args.force_venv_backend or noxfile_args.force_venv_backend
+        return "none"
+    return command_args.force_venv_backend or noxfile_args.force_venv_backend  # type: ignore[no-any-return]
 
 
 def _envdir_merge_func(
@@ -137,11 +149,45 @@ def _envdir_merge_func(
     return command_args.envdir or noxfile_args.envdir or ".nox"
 
 
-def _sessions_default() -> Optional[List[str]]:
-    """Looks at the NOXSESSION env var to set the default value for sessions."""
-    nox_env = os.environ.get("NOXSESSION")
-    env_sessions = nox_env.split(",") if nox_env else None
-    return env_sessions
+def _reuse_venv_merge_func(
+    command_args: argparse.Namespace, noxfile_args: argparse.Namespace
+) -> ReuseVenvType:
+    """Merge reuse_venv from command args and Noxfile while maintaining
+    backwards compatibility with reuse_existing_virtualenvs. Default is "no".
+
+    Args:
+        command_args (_option_set.Namespace): The options specified on the
+            command-line.
+        noxfile_Args (_option_set.Namespace): The options specified in the
+            Noxfile.
+    """
+    # back-compat scenario with no_reuse_existing_virtualenvs/reuse_existing_virtualenvs
+    if command_args.no_reuse_existing_virtualenvs:
+        return "no"
+    if (
+        command_args.reuse_existing_virtualenvs
+        or noxfile_args.reuse_existing_virtualenvs
+    ):
+        return "yes"
+    # regular option behavior
+    return command_args.reuse_venv or noxfile_args.reuse_venv or "no"
+
+
+def default_env_var_list_factory(env_var: str) -> Callable[[], list[str] | None]:
+    """Looks at the env var to set the default value for a list of env vars.
+
+    Args:
+        env_var (str): The name of the environment variable to look up.
+
+    Returns:
+        A callback that retrieves a list from a comma-delimited environment variable.
+    """
+
+    def _default_list() -> list[str] | None:
+        env_value = os.environ.get(env_var)
+        return env_value.split(",") if env_value else None
+
+    return _default_list
 
 
 def _color_finalizer(value: bool, args: argparse.Namespace) -> bool:
@@ -154,15 +200,15 @@ def _color_finalizer(value: bool, args: argparse.Namespace) -> bool:
     Returns:
         The new value for the "color" option.
     """
-    if args.forcecolor is True and args.nocolor is True:
+    if args.forcecolor and args.nocolor:
         raise _option_set.ArgumentError(
             None, "Can not specify both --no-color and --force-color."
         )
 
-    if args.forcecolor is True:
+    if args.forcecolor:
         return True
 
-    if args.nocolor is True:
+    if args.nocolor or "NO_COLOR" in os.environ:
         return False
 
     return sys.stdout.isatty()
@@ -178,15 +224,25 @@ def _force_pythons_finalizer(
 
 
 def _R_finalizer(value: bool, args: argparse.Namespace) -> bool:
-    """Propagate -R to --reuse-existing-virtualenvs and --no-install."""
+    """Propagate -R to --reuse-existing-virtualenvs and --no-install and --reuse-venv=yes."""
     if value:
+        args.reuse_venv = "yes"
         args.reuse_existing_virtualenvs = args.no_install = value
     return value
 
 
+def _reuse_existing_virtualenvs_finalizer(
+    value: bool, args: argparse.Namespace
+) -> bool:
+    """Propagate --reuse-existing-virtualenvs to --reuse-venv=yes."""
+    if value:
+        args.reuse_venv = "yes"
+    return value
+
+
 def _posargs_finalizer(
     value: Sequence[Any], args: argparse.Namespace
-) -> Union[Sequence[Any], List[Any]]:
+) -> Sequence[Any] | list[Any]:
     """Removes the leading "--"s in the posargs array (if any) and asserts that
     remaining arguments came after a "--".
     """
@@ -197,31 +253,55 @@ def _posargs_finalizer(
     if "--" not in posargs:
         unexpected_posargs = posargs
         raise _option_set.ArgumentError(
-            None, "Unknown argument(s) '{}'.".format(" ".join(unexpected_posargs))
+            None, f"Unknown argument(s) '{' '.join(unexpected_posargs)}'."
         )
 
     dash_index = posargs.index("--")
     if dash_index != 0:
         unexpected_posargs = posargs[0:dash_index]
         raise _option_set.ArgumentError(
-            None, "Unknown argument(s) '{}'.".format(" ".join(unexpected_posargs))
+            None, f"Unknown argument(s) '{' '.join(unexpected_posargs)}'."
         )
 
     return posargs[dash_index + 1 :]
 
 
+def _python_completer(
+    prefix: str, parsed_args: argparse.Namespace, **kwargs: Any
+) -> Iterable[str]:
+    module = load_nox_module(parsed_args)
+    manifest = discover_manifest(module, parsed_args)
+    return filter(
+        None,
+        (
+            session.func.python  # type:ignore[misc] # str sequences flattened, other non-strs falsey and filtered out
+            for session, _ in manifest.list_all_sessions()
+        ),
+    )
+
+
 def _session_completer(
     prefix: str, parsed_args: argparse.Namespace, **kwargs: Any
-) -> List[str]:
-    global_config = parsed_args
-    module = load_nox_module(global_config)
-    manifest = discover_manifest(module, global_config)
-    filtered_manifest = filter_manifest(manifest, global_config)
+) -> Iterable[str]:
+    parsed_args.list_sessions = True
+    module = load_nox_module(parsed_args)
+    manifest = discover_manifest(module, parsed_args)
+    filtered_manifest = filter_manifest(manifest, parsed_args)
     if isinstance(filtered_manifest, int):
         return []
-    return [
+    return (
         session.friendly_name for session, _ in filtered_manifest.list_all_sessions()
-    ]
+    )
+
+
+def _tag_completer(
+    prefix: str, parsed_args: argparse.Namespace, **kwargs: Any
+) -> Iterable[str]:
+    module = load_nox_module(parsed_args)
+    manifest = discover_manifest(module, parsed_args)
+    return itertools.chain.from_iterable(
+        filter(None, (session.tags for session, _ in manifest.list_all_sessions()))
+    )
 
 
 options.add_options(
@@ -249,6 +329,13 @@ options.add_options(
         action="store_true",
         help="List all available sessions and exit.",
     ),
+    _option_set.Option(
+        "json",
+        "--json",
+        group=options.groups["sessions"],
+        action="store_true",
+        help="JSON output formatting. Requires list-sessions currently.",
+    ),
     _option_set.Option(
         "sessions",
         "-s",
@@ -257,9 +344,9 @@ options.add_options(
         "--session",
         group=options.groups["sessions"],
         noxfile=True,
-        merge_func=functools.partial(_sessions_and_keywords_merge_func, "sessions"),
+        merge_func=functools.partial(_sessions_merge_func, "sessions"),
         nargs="*",
-        default=_sessions_default,
+        default=default_env_var_list_factory("NOXSESSION"),
         help="Which sessions to run. By default, all sessions will run.",
         completer=_session_completer,
     ),
@@ -271,7 +358,9 @@ options.add_options(
         group=options.groups["python"],
         noxfile=True,
         nargs="*",
+        default=default_env_var_list_factory("NOXPYTHON"),
         help="Only run sessions that use the given python interpreter versions.",
+        completer=_python_completer,
     ),
     _option_set.Option(
         "keywords",
@@ -279,8 +368,20 @@ options.add_options(
         "--keywords",
         group=options.groups["sessions"],
         noxfile=True,
-        merge_func=functools.partial(_sessions_and_keywords_merge_func, "keywords"),
+        merge_func=functools.partial(_sessions_merge_func, "keywords"),
         help="Only run sessions that match the given expression.",
+        completer=argcomplete.completers.ChoicesCompleter(()),
+    ),
+    _option_set.Option(
+        "tags",
+        "-t",
+        "--tags",
+        group=options.groups["sessions"],
+        noxfile=True,
+        merge_func=functools.partial(_sessions_merge_func, "tags"),
+        nargs="*",
+        help="Only run sessions with the given tags.",
+        completer=_tag_completer,
     ),
     _option_set.Option(
         "posargs",
@@ -306,7 +407,6 @@ options.add_options(
         group=options.groups["reporting"],
         action="store_true",
         help="Adds a timestamp to logged output.",
-        noxfile=True,
     ),
     _option_set.Option(
         "default_venv_backend",
@@ -314,10 +414,13 @@ options.add_options(
         "--default-venv-backend",
         group=options.groups["environment"],
         noxfile=True,
+        default=lambda: os.environ.get("NOX_DEFAULT_VENV_BACKEND"),
         merge_func=_default_venv_backend_merge_func,
-        help="Virtual environment backend to use by default for nox sessions, this is ``'virtualenv'`` by default but "
-        "any of ``('virtualenv', 'conda', 'venv')`` are accepted.",
-        choices=["none", "virtualenv", "conda", "venv"],
+        help=(
+            "Virtual environment backend to use by default for Nox sessions, this is"
+            f" ``'virtualenv'`` by default but any of ``{list(ALL_VENVS)!r}`` are accepted."
+        ),
+        choices=list(ALL_VENVS),
     ),
     _option_set.Option(
         "force_venv_backend",
@@ -326,10 +429,12 @@ options.add_options(
         group=options.groups["environment"],
         noxfile=True,
         merge_func=_force_venv_backend_merge_func,
-        help="Virtual environment backend to force-use for all nox sessions in this run, overriding any other venv "
-        "backend declared in the nox file and ignoring the default backend. Any of ``('virtualenv', 'conda', 'venv')`` "
-        "are accepted.",
-        choices=["none", "virtualenv", "conda", "venv"],
+        help=(
+            "Virtual environment backend to force-use for all Nox sessions in this run,"
+            " overriding any other venv backend declared in the Noxfile and ignoring"
+            f" the default backend. Any of ``{list(ALL_VENVS)!r}`` are accepted."
+        ),
+        choices=list(ALL_VENVS),
     ),
     _option_set.Option(
         "no_venv",
@@ -337,15 +442,33 @@ options.add_options(
         group=options.groups["environment"],
         default=False,
         action="store_true",
-        help="Runs the selected sessions directly on the current interpreter, without creating a venv. This is an alias "
-        "for '--force-venv-backend none'.",
+        help=(
+            "Runs the selected sessions directly on the current interpreter, without"
+            " creating a venv. This is an alias for '--force-venv-backend none'."
+        ),
+    ),
+    _option_set.Option(
+        "reuse_venv",
+        "--reuse-venv",
+        group=options.groups["environment"],
+        noxfile=True,
+        merge_func=_reuse_venv_merge_func,
+        help=(
+            "Controls existing virtualenvs recreation. This is ``'no'`` by"
+            " default, but any of ``('yes', 'no', 'always', 'never')`` are accepted."
+        ),
+        choices=["yes", "no", "always", "never"],
     ),
     *_option_set.make_flag_pair(
         "reuse_existing_virtualenvs",
         ("-r", "--reuse-existing-virtualenvs"),
-        ("--no-reuse-existing-virtualenvs",),
+        (
+            "-N",
+            "--no-reuse-existing-virtualenvs",
+        ),
         group=options.groups["environment"],
-        help="Re-use existing virtualenvs instead of recreating them.",
+        help="This is an alias for '--reuse-venv=yes|no'.",
+        finalizer_func=_reuse_existing_virtualenvs_finalizer,
     ),
     _option_set.Option(
         "R",
@@ -354,7 +477,7 @@ options.add_options(
         group=options.groups["environment"],
         action="store_true",
         help=(
-            "Re-use existing virtualenvs and skip package re-installation."
+            "Reuse existing virtualenvs and skip package re-installation."
             " This is an alias for '--reuse-existing-virtualenvs --no-install'."
         ),
         finalizer_func=_R_finalizer,
@@ -365,7 +488,7 @@ options.add_options(
         "--noxfile",
         group=options.groups["general"],
         default="noxfile.py",
-        help="Location of the Python file containing nox sessions.",
+        help="Location of the Python file containing Nox sessions.",
     ),
     _option_set.Option(
         "envdir",
@@ -373,7 +496,8 @@ options.add_options(
         noxfile=True,
         merge_func=_envdir_merge_func,
         group=options.groups["environment"],
-        help="Directory where nox will store virtualenvs, this is ``.nox`` by default.",
+        help="Directory where Nox will store virtualenvs, this is ``.nox`` by default.",
+        completer=argcomplete.completers.DirectoriesCompleter(),
     ),
     _option_set.Option(
         "extra_pythons",
@@ -381,19 +505,25 @@ options.add_options(
         "--extra-python",
         group=options.groups["python"],
         nargs="*",
+        default=default_env_var_list_factory("NOXEXTRAPYTHON"),
         help="Additionally, run sessions using the given python interpreter versions.",
+        completer=_python_completer,
     ),
     _option_set.Option(
         "force_pythons",
+        "-P",
         "--force-pythons",
         "--force-python",
         group=options.groups["python"],
         nargs="*",
+        default=default_env_var_list_factory("NOXFORCEPYTHON"),
         help=(
-            "Run sessions with the given interpreters instead of those listed in the Noxfile."
-            " This is a shorthand for ``--python=X.Y --extra-python=X.Y``."
+            "Run sessions with the given interpreters instead of those listed in the"
+            " Noxfile. This is a shorthand for ``--python=X.Y --extra-python=X.Y``."
+            " It will also work on sessions that don't have any interpreter parametrized."
         ),
         finalizer_func=_force_pythons_finalizer,
+        completer=_python_completer,
     ),
     *_option_set.make_flag_pair(
         "stop_on_first_error",
@@ -408,13 +538,17 @@ options.add_options(
         ("--no-error-on-missing-interpreters",),
         group=options.groups["execution"],
         help="Error instead of skipping sessions if an interpreter can not be located.",
+        default=lambda: "CI" in os.environ,
     ),
     *_option_set.make_flag_pair(
         "error_on_external_run",
         ("--error-on-external-run",),
         ("--no-error-on-external-run",),
         group=options.groups["execution"],
-        help="Error if run() is used to execute a program that isn't installed in a session's virtualenv.",
+        help=(
+            "Error if run() is used to execute a program that isn't installed in a"
+            " session's virtualenv."
+        ),
     ),
     _option_set.Option(
         "install_only",
@@ -431,7 +565,7 @@ options.add_options(
         action="store_true",
         help=(
             "Skip invocations of session methods for installing packages"
-            " (session.install, session.conda_install, session.run_always)"
+            " (session.install, session.conda_install, session.run_install)"
             " when a virtualenv is being reused."
         ),
     ),
@@ -441,13 +575,17 @@ options.add_options(
         group=options.groups["reporting"],
         noxfile=True,
         help="Output a report of all sessions to the given filename.",
+        completer=argcomplete.completers.FilesCompleter(("json",)),
     ),
     _option_set.Option(
         "non_interactive",
         "--non-interactive",
         group=options.groups["execution"],
         action="store_true",
-        help="Force session.interactive to always be False, even in interactive sessions.",
+        help=(
+            "Force session.interactive to always be False, even in interactive"
+            " sessions."
+        ),
     ),
     _option_set.Option(
         "nocolor",
@@ -463,7 +601,7 @@ options.add_options(
         "--forcecolor",
         "--force-color",
         group=options.groups["reporting"],
-        default=False,
+        default=lambda: "FORCE_COLOR" in os.environ,
         action="store_true",
         help="Force color output, even if stdout is not an interactive terminal.",
     ),
@@ -474,6 +612,14 @@ options.add_options(
         hidden=True,
         finalizer_func=_color_finalizer,
     ),
+    # Stores the original working directory that Nox was invoked from,
+    # since it could be different from the Noxfile's directory.
+    _option_set.Option(
+        "invoked_from",
+        group=None,
+        hidden=True,
+        default=os.getcwd,
+    ),
 )
 
 
