@@ -16,16 +16,11 @@ from __future__ import annotations
 
 import ast
 import contextlib
-import sys
+from importlib import metadata
 
 from packaging.specifiers import InvalidSpecifier, SpecifierSet
 from packaging.version import InvalidVersion, Version
 
-if sys.version_info >= (3, 8):  # pragma: no cover
-    import importlib.metadata as metadata
-else:  # pragma: no cover
-    import importlib_metadata as metadata
-
 
 class VersionCheckFailed(Exception):
     """The Nox version does not satisfy what ``nox.needs_version`` specifies."""
@@ -42,10 +37,7 @@ def get_nox_version() -> str:
 
 def _parse_string_constant(node: ast.AST) -> str | None:  # pragma: no cover
     """Return the value of a string constant."""
-    if sys.version_info < (3, 8):
-        if isinstance(node, ast.Str) and isinstance(node.s, str):
-            return node.s
-    elif isinstance(node, ast.Constant) and isinstance(node.value, str):
+    if isinstance(node, ast.Constant) and isinstance(node.value, str):
         return node.value
     return None
 
